"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.LowLevelNetwork = void 0;
const fs_1 = require("fs");
const net_1 = require("net");
const path_1 = __importDefault(require("path"));
const atvik_1 = require("atvik");
const debug_1 = __importDefault(require("debug"));
const proper_lockfile_1 = require("proper-lockfile");
/**
 * Resolve the given path to a path that can be used with an IPC socket.
 *
 * @param p -
 *   path to resolve
 * @returns
 *   resolved path
 */
function toSocketPath(p) {
    const resolved = path_1.default.resolve(p);
    if (process.platform === 'win32') {
        // On Windows IPC sockets need to be either in \\?\pipe\ or \\.\pipe\
        return path_1.default.join('\\\\?\\pipe', resolved);
    }
    else {
        // Other platforms are assumed to be Unix-like where sockets can be created anywhere
        return resolved;
    }
}
/**
 * Generate a random retry time between 30 and 130 ms.
 *
 * @returns
 *   time in milliseconds
 */
function randomRetryTime() {
    return Math.floor(30 + (Math.random() * 100));
}
const client = Symbol('client');
const server = Symbol('server');
const releaseLock = Symbol('releaseLock');
const readyEvent = Symbol('readyEvent');
const errorEvent = Symbol('errorEvent');
const leaderEvent = Symbol('leaderEvent');
const connectEvent = Symbol('connectEvent');
const connectionEvent = Symbol('connectionEvent');
/**
 * Low-level network that provides sockets to and from a leader. When an
 * instance of this network is connected it can either become the leader, in
 * which case a server socket will be opened, or a client in which it will
 * connect to the leader.
 *
 * When connected the network will attempt to stay connected.
 */
class LowLevelNetwork {
    constructor(options) {
        if (typeof options !== 'object') {
            throw new Error('Options are needed to create a new network');
        }
        if (typeof options.path !== 'string') {
            throw new Error('Path for socket is required');
        }
        this.options = {
            path: options.path
        };
        this.debug = debug_1.default('local-machine-network:' + (options.id || ('[' + options.path + ']')));
        this.leader = false;
        this.started = false;
        this[readyEvent] = new atvik_1.Event(this);
        this[errorEvent] = new atvik_1.Event(this);
        this[connectEvent] = new atvik_1.Event(this);
        this[connectionEvent] = new atvik_1.Event(this);
        this[leaderEvent] = new atvik_1.Event(this);
    }
    get onConnect() {
        return this[connectEvent].subscribable;
    }
    get onConnection() {
        return this[connectionEvent].subscribable;
    }
    /**
     * Event emitted when this instance becomes the leader of the local
     * network.
     *
     * @returns
     *   subscribable
     */
    get onLeader() {
        return this[leaderEvent].subscribable;
    }
    /**
     * Event emitted when this instance becomes ready.
     *
     * @returns
     *   subscribable
     */
    get onReady() {
        return this[readyEvent].subscribable;
    }
    /**
     * Event emitted if an error occurs.
     *
     * @returns
     *   subscribable
     */
    get onError() {
        return this[errorEvent].subscribable;
    }
    /**
     * Start this network.
     */
    start() {
        if (this.started) {
            // TODO: If we are started return promise that resolves if connected
        }
        this.started = true;
        return new Promise((resolve, reject) => {
            const readyHandle = this.onReady(() => {
                readyHandle.unsubscribe();
                errorHandle.unsubscribe();
                resolve();
            });
            const errorHandle = this.onError(() => {
                readyHandle.unsubscribe();
                errorHandle.unsubscribe();
                reject();
            });
            // Attempt to connect or bind
            this.attemptConnectOrBind();
        });
    }
    /**
     * Attempt to connect or become the master.
     *
     * The flow here is:
     *
     * 1) Attempt to acquire lock
     * 2a) If lock acquired: Bind and become the master
     * 2b): Otherwise: Connect to the master
     */
    attemptConnectOrBind() {
        // Resolve the paths to work with
        const socketPath = toSocketPath(this.options.path);
        // Reset leader flag
        this.leader = false;
        /*
         * Attempt to get a lock, hopefully allowing us to create the
         * socket and become the leader.
         */
        proper_lockfile_1.lock(this.options.path, {
            /*
             * As the socket doesn't exist before we lock we don't use realpath
             * and instead let lock be created wherever.
             */
            realpath: false,
            onCompromised: () => {
                const release = this[releaseLock];
                if (release) {
                    this.debug('Lock on server path compromised, attempting release');
                    release()
                        .catch(err => this.debug('Unable to unlock, error: ', err))
                        .then(() => {
                        this[releaseLock] = undefined;
                        this.reconnect('Attempting reconnect after compromised lock');
                    });
                }
                else {
                    this.reconnect('Lock on server path compromised');
                }
            }
        })
            .then(release => {
            /**
             * We now have the lock, try creating a server.
             */
            // Store a reference to be able to release the lock
            this[releaseLock] = release;
            this.attemptBind(socketPath);
        })
            .catch(() => {
            /*
             * Someone else is already holding the lock - try to
             * connect to them.
             */
            this.attemptConnect(socketPath);
        });
    }
    /**
     * Unsafe: Release the lock if we have it. This is done if the server
     * fails to bind or if the network is disconnected.
     *
     * @returns
     *   promise that resolves when the lock is released
     */
    _releaseLock() {
        const release = this[releaseLock];
        if (!release)
            return Promise.resolve();
        this[releaseLock] = undefined;
        return release();
    }
    reconnect(message) {
        // Check that this network is started
        if (!this.started)
            return;
        // Resolve when to next retry a connection
        const retryTime = randomRetryTime();
        this.debug(message + '; Retrying in ' + retryTime + ' ms');
        setTimeout(() => this.attemptConnectOrBind(), retryTime);
    }
    attemptConnect(socketPath) {
        this.debug('Connecting as client');
        const socket = net_1.connect(socketPath, () => {
            this.debug('Connected to leader');
            this[client] = socket;
            this[connectEvent].emit(socket);
            // Indicate that the network is ready
            this[readyEvent].emit();
        });
        socket.on('error', err => {
            // Mark the client as non-existent
            const hadConnected = !!this[client];
            this[client] = undefined;
            // Resolve a readable debug message
            let message;
            const code = typeof err.code !== 'undefined' ? err.code : 'Unknown Error';
            if (hadConnected) {
                message = 'Disconnected from server with ' + code;
            }
            else {
                message = 'Could not connect to server, got code ' + code;
            }
            this.reconnect(message);
        });
        socket.on('close', hadError => {
            if (hadError)
                return;
            // Mark the client as non-existent
            this[client] = undefined;
            // Request a reconnect
            this.reconnect('Disconnected from server without error');
        });
    }
    attemptBind(socketPath, attempt = 0) {
        if (attempt > 10) {
            // Already tried to bind 10 times - just give up at this point
            this.debug('Could not bind server socket, giving up after 10 tries');
            this[errorEvent].emit(new Error('Could not bind server socket, giving up after 10 tries'));
            return;
        }
        this.debug('Creating socket as leader');
        // Try to unlink the socket path
        fs_1.unlink(socketPath, unlinkErr => {
            if (unlinkErr && unlinkErr.code !== 'ENOENT') {
                const retryTime = randomRetryTime();
                this.debug('Could not unlink existing socket, errored with: ' + unlinkErr.code + '; Retrying in', retryTime, 'ms');
                setTimeout(() => this.attemptBind(socketPath, attempt + 1), retryTime);
                return;
            }
            const serverSocket = net_1.createServer();
            serverSocket.listen(socketPath, () => {
                // Server was successfully bound, store reference and resolve callback
                this.debug('Hosting network as leader');
                this[server] = serverSocket;
                // Mark as leader and emit event
                this.leader = true;
                this[leaderEvent].emit(serverSocket);
                // Indicate that the network is ready
                this[readyEvent].emit();
            });
            serverSocket.on('connection', socket => {
                this[connectionEvent].emit(socket);
            });
            serverSocket.on('error', err => {
                // Something went wrong, close this server and reconnect
                // Mark the server as non-existent
                const hadBound = !!this[server];
                this[server] = undefined;
                if (hadBound) {
                    // Make sure the socket is closed
                    serverSocket.close();
                    this.reconnect('Server errored with ' + err.code);
                }
                else {
                    const retryTime = randomRetryTime();
                    this.debug('Could not bind socket, errored with ' + err.code + '; Retrying in', retryTime, 'ms');
                    setTimeout(() => this.attemptBind(socketPath, attempt + 1), retryTime);
                }
            });
        });
    }
    stop() {
        this.started = false;
        const serverSocket = this[server];
        if (serverSocket) {
            return new Promise((resolve, reject) => {
                serverSocket.close(() => {
                    // First unlink the socket path
                    const socketPath = toSocketPath(this.options.path);
                    fs_1.unlink(socketPath, () => {
                        this._releaseLock()
                            .then(resolve)
                            .catch(reject);
                    });
                });
            });
        }
        else {
            const clientSocket = this[client];
            if (clientSocket) {
                return new Promise(resolve => {
                    clientSocket.destroy();
                    resolve();
                });
            }
            else {
                return Promise.resolve();
            }
        }
    }
}
exports.LowLevelNetwork = LowLevelNetwork;
//# sourceMappingURL=LowLevelNetwork.js.map