"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ObjectSocket = exports.ObjectNetwork = void 0;
const atvik_1 = require("atvik");
const debug_1 = __importDefault(require("debug"));
const LowLevelNetwork_1 = require("../LowLevelNetwork");
const packets_1 = require("./packets");
const connectEvent = Symbol('connectEvent');
const disconnectEvent = Symbol('disconnectEvent');
const leaderEvent = Symbol('leaderEvent');
const connectionEvent = Symbol('connectionEvent');
const messageEvent = Symbol('messageEvent');
/**
 * Adapter over a low-level network for those cases where clients want to
 * exchange objects with the leader. This implementation will use a codec
 * to transform objects to and from a binary format sent between the peers of
 * the network.
 */
class ObjectNetwork {
    /**
     * Create a new object network.
     *
     * @param options -
     *   options for the network
     */
    constructor(options) {
        if (!options) {
            throw new Error('options are required');
        }
        if (!options.path) {
            throw new Error('path needs to be specified');
        }
        if (!options.codec) {
            throw new Error('codec needs to be specified');
        }
        this.debug = debug_1.default('local-machine-network:' + (options.id || ('[' + options.path + ']')));
        // Create the low-level network
        this.lowLevel = new LowLevelNetwork_1.LowLevelNetwork(options);
        this.codec = options.codec;
        this[connectEvent] = new atvik_1.Event(this);
        this[connectionEvent] = new atvik_1.Event(this);
        this[leaderEvent] = new atvik_1.Event(this);
        this[messageEvent] = new atvik_1.Event(this);
        /*
         * When this instance becomes the leader make sure to intercept
         * server socket errors and to emit the leader event.
         */
        this.lowLevel.onLeader(serverSocket => {
            // Listen to all socket errors to avoid unexpected exits
            serverSocket.on('error', err => {
                this.debug('Caught error from server socket', err);
            });
            // If previously connected to a leader remove the reference
            this.connection = undefined;
            // Emit the leader event
            this[leaderEvent].emit();
        });
        /*
         * On incoming connections make sure to wrap them with encoding and
         * decoding.
         */
        this.lowLevel.onConnection(clientSocket => {
            // Create the wrapped socket
            const connection = new ObjectSocket({
                socket: clientSocket,
                debug: this.debug,
                emitMessage: msg => this[messageEvent].emit(msg),
                send: this.send.bind(this),
                codec: this.codec
            });
            this[connectionEvent].emit(connection);
        });
        /*
         * When this instance is not the leader and it connects to something
         * save a reference to the connection.
         */
        this.lowLevel.onConnect(socket => {
            const connection = new ObjectSocket({
                socket: socket,
                debug: this.debug,
                emitMessage: msg => this[messageEvent].emit(msg),
                send: this.send.bind(this),
                codec: this.codec
            });
            this.connection = connection;
            this[connectEvent].emit(connection);
        });
    }
    get onConnect() {
        return this[connectEvent].subscribable;
    }
    get onLeader() {
        return this[leaderEvent].subscribable;
    }
    get onConnection() {
        return this[connectionEvent].subscribable;
    }
    get onMessage() {
        return this[messageEvent].subscribable;
    }
    /**
     * Connect to the network.
     *
     * @returns
     *   promise that resolves when network is started
     */
    start() {
        return this.lowLevel.start();
    }
    /**
     * Disconnect from the network.
     *
     * @returns
     *   promise that resolve when the network stops
     */
    stop() {
        return this.lowLevel.stop();
    }
    /**
     * Send a message to the current leader.
     *
     * @param message -
     *   message to send
     */
    send(message) {
        if (this.lowLevel.leader) {
            // This instance is the leader, emit the event locally
            this[messageEvent].emit({
                returnPath: this,
                data: message
            });
        }
        else if (this.connection) {
            // Connected to the leader send via the active socket
            this.connection.send(message);
        }
        else {
            this.debug('Unable to send message, not connected, message was:', message);
            throw new Error('Unable to send message, not connected');
        }
    }
}
exports.ObjectNetwork = ObjectNetwork;
class ObjectSocket {
    constructor(control) {
        this.control = control;
        this[disconnectEvent] = new atvik_1.Event(this);
        this[messageEvent] = new atvik_1.Event(this);
        // Intercept errors and output them as debug info
        control.socket.on('error', err => {
            control.debug('Error received:', err);
        });
        // Emit disconnection events
        control.socket.on('close', () => this[disconnectEvent].emit());
        // Setup the decoder for incoming messages
        const decoder = new packets_1.PacketDecodingStream();
        const pipe = control.socket.pipe(decoder);
        // When data is received emit an event
        decoder.on('data', data => {
            const msg = control.codec.decode(data);
            control.debug('Received message:', msg);
            const message = {
                returnPath: this,
                data: msg
            };
            control.emitMessage(message);
            this[messageEvent].emit(message);
        });
        // Catch errors on pipe and decoder
        decoder.on('error', err => control.debug('Error from decoder', err));
        pipe.on('error', err => control.debug('Error from pipe', err));
    }
    get onDisconnect() {
        return this[disconnectEvent].subscribable;
    }
    get onMessage() {
        return this[messageEvent].subscribable;
    }
    /**
     * Send a message to this instance.
     *
     * @param message -
     *   message to send
     */
    send(message) {
        const data = this.control.codec.encode(message);
        const packet = packets_1.encodePacket(data);
        try {
            this.control.debug('Sending message to leader:', message);
            this.control.socket.write(packet);
        }
        catch (err) {
            this.control.debug('Could not write, got an error:', err);
        }
    }
}
exports.ObjectSocket = ObjectSocket;
//# sourceMappingURL=ObjectNetwork.js.map