/// <reference types="node" />
import { Socket, Server } from 'net';
/**
 * Options for creating a new low level network.
 */
export interface LowLevelNetworkOptions {
    /**
     * Optional id used for debug messages.
     */
    id?: string;
    /**
     * Path at which the network should create its socket.
     */
    path: string;
}
declare const client: unique symbol;
declare const server: unique symbol;
declare const releaseLock: unique symbol;
declare const readyEvent: unique symbol;
declare const errorEvent: unique symbol;
declare const leaderEvent: unique symbol;
declare const connectEvent: unique symbol;
declare const connectionEvent: unique symbol;
/**
 * Low-level network that provides sockets to and from a leader. When an
 * instance of this network is connected it can either become the leader, in
 * which case a server socket will be opened, or a client in which it will
 * connect to the leader.
 *
 * When connected the network will attempt to stay connected.
 */
export declare class LowLevelNetwork {
    private debug;
    private options;
    started: boolean;
    leader: boolean;
    private [client]?;
    private [server]?;
    private [releaseLock]?;
    private [readyEvent];
    private [errorEvent];
    private [connectEvent];
    private [connectionEvent];
    private [leaderEvent];
    constructor(options: LowLevelNetworkOptions);
    get onConnect(): import("atvik").Subscribable<this, [Socket]>;
    get onConnection(): import("atvik").Subscribable<this, [Socket]>;
    /**
     * Event emitted when this instance becomes the leader of the local
     * network.
     *
     * @returns
     *   subscribable
     */
    get onLeader(): import("atvik").Subscribable<this, [Server]>;
    /**
     * Event emitted when this instance becomes ready.
     *
     * @returns
     *   subscribable
     */
    get onReady(): import("atvik").Subscribable<this, []>;
    /**
     * Event emitted if an error occurs.
     *
     * @returns
     *   subscribable
     */
    get onError(): import("atvik").Subscribable<this, [Error]>;
    /**
     * Start this network.
     */
    start(): Promise<void>;
    /**
     * Attempt to connect or become the master.
     *
     * The flow here is:
     *
     * 1) Attempt to acquire lock
     * 2a) If lock acquired: Bind and become the master
     * 2b): Otherwise: Connect to the master
     */
    private attemptConnectOrBind;
    /**
     * Unsafe: Release the lock if we have it. This is done if the server
     * fails to bind or if the network is disconnected.
     *
     * @returns
     *   promise that resolves when the lock is released
     */
    private _releaseLock;
    private reconnect;
    private attemptConnect;
    private attemptBind;
    stop(): Promise<void>;
}
export {};
