/// <reference types="node" />
import { Socket } from 'net';
import debug from 'debug';
import { ObjectCodec } from './ObjectCodec';
declare const connectEvent: unique symbol;
declare const disconnectEvent: unique symbol;
declare const leaderEvent: unique symbol;
declare const connectionEvent: unique symbol;
declare const messageEvent: unique symbol;
/**
 * Options for creating an object network.
 */
export interface ObjectNetworkOptions {
    /**
     * Optional id used for debug messages.
     */
    id?: string;
    /**
     * Path at which the network should create its socket.
     */
    path: string;
    /**
     * The codec that is used for messages.
     */
    codec: ObjectCodec;
}
/**
 * Interface for things that can send an object message.
 */
export interface ObjectSender {
    send(message: any): void;
}
/**
 * Message that can be received via the object network.
 */
export interface ObjectMessage {
    /**
     * Return path of the message, can be used to send a message to whoever
     * sent this message.
     */
    returnPath: ObjectSender;
    /**
     * Data carried by message.
     */
    data: any;
}
/**
 * Adapter over a low-level network for those cases where clients want to
 * exchange objects with the leader. This implementation will use a codec
 * to transform objects to and from a binary format sent between the peers of
 * the network.
 */
export declare class ObjectNetwork {
    private debug;
    private lowLevel;
    private connection?;
    private readonly codec;
    private [connectEvent];
    private [connectionEvent];
    private [leaderEvent];
    private [messageEvent];
    /**
     * Create a new object network.
     *
     * @param options -
     *   options for the network
     */
    constructor(options: ObjectNetworkOptions);
    get onConnect(): import("atvik").Subscribable<this, [ObjectSocket]>;
    get onLeader(): import("atvik").Subscribable<this, []>;
    get onConnection(): import("atvik").Subscribable<this, [ObjectSocket]>;
    get onMessage(): import("atvik").Subscribable<this, [ObjectMessage]>;
    /**
     * Connect to the network.
     *
     * @returns
     *   promise that resolves when network is started
     */
    start(): Promise<void>;
    /**
     * Disconnect from the network.
     *
     * @returns
     *   promise that resolve when the network stops
     */
    stop(): Promise<void>;
    /**
     * Send a message to the current leader.
     *
     * @param message -
     *   message to send
     */
    send(message: any): void;
}
interface ObjectSocketControl {
    send(message: any): void;
    emitMessage(message: ObjectMessage): void;
    debug: debug.Debugger;
    socket: Socket;
    codec: ObjectCodec;
}
export declare class ObjectSocket {
    private control;
    private [disconnectEvent];
    private [messageEvent];
    constructor(control: ObjectSocketControl);
    get onDisconnect(): import("atvik").Subscribable<this, []>;
    get onMessage(): import("atvik").Subscribable<this, [ObjectMessage]>;
    /**
     * Send a message to this instance.
     *
     * @param message -
     *   message to send
     */
    send(message: any): void;
}
export {};
